from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any

MODULE_ATTRS = {
    "Resource": "._internal.resource:Resource",
    "Runnable": "._internal.runner:Runnable",
    "Runner": "._internal.runner:Runner",
    "Strategy": "._internal.runner.strategy:Strategy",
    "Service": "._internal.service:Service",
    "HTTPServer": ".server:HTTPServer",
    "GrpcServer": ".server:GrpcServer",
}

__all__ = [
    "Resource",
    "Runnable",
    "Runner",
    "Strategy",
    "Service",
    "HTTPServer",
    "GrpcServer",
]

if TYPE_CHECKING:
    from ._internal.resource import Resource
    from ._internal.runner import Runnable
    from ._internal.runner import Runner
    from ._internal.runner.strategy import Strategy
    from ._internal.service import Service
    from .server import GrpcServer
    from .server import HTTPServer

else:

    def __getattr__(name: str) -> Any:
        if name in MODULE_ATTRS:
            from importlib import import_module

            module_name, attr_name = MODULE_ATTRS[name].split(":")
            module = import_module(module_name, __package__)
            return getattr(module, attr_name)
        raise AttributeError(f"module {__name__} has no attribute {name}")



from __future__ import annotations

from typing import TYPE_CHECKING

from .schemas import BentoBuildEvent

if TYPE_CHECKING:
    from ...bento.bento import Bento


def _cli_bentoml_build_event(
    cmd_group: str,
    cmd_name: str,
    return_value: Bento | None,
) -> BentoBuildEvent:  # pragma: no cover
    from ...bento.bento import BentoInfo
    from ...bento.bento import BentoInfoV2

    if return_value is None:
        return BentoBuildEvent(cmd_group=cmd_group, cmd_name=cmd_name)
    bento = return_value
    total_size = bento.total_size()
    if isinstance(bento.info, BentoInfoV2):
        num_of_runners = len(bento.info.services) - 1
    elif isinstance(bento.info, BentoInfo):
        num_of_runners = len(bento.info.runners)
    else:
        num_of_runners = 0
    return BentoBuildEvent(
        cmd_group=cmd_group,
        cmd_name=cmd_name,
        bento_creation_timestamp=bento.info.creation_time,
        bento_size_in_kb=bento.file_size / 1024,
        model_size_in_kb=(total_size - bento.file_size) / 1024,
        num_of_models=len(bento.info.all_models),
        num_of_runners=num_of_runners,
        model_types=[m.module or "" for m in bento.info.all_models],
    )


cli_events_map = {"bentos": {"build": _cli_bentoml_build_event}}



from __future__ import annotations

import hashlib
import logging
import platform
import subprocess
import sys
import typing as t
from pathlib import Path

import attrs
import fs

from bentoml._internal.bento.bento import ImageInfo
from bentoml._internal.bento.build_config import BentoBuildConfig
from bentoml._internal.configuration import DEFAULT_LOCK_PLATFORM
from bentoml._internal.configuration import get_bentoml_requirement
from bentoml._internal.configuration import get_debug_mode
from bentoml._internal.configuration import get_quiet_mode
from bentoml._internal.container.frontend.dockerfile import CONTAINER_METADATA
from bentoml._internal.container.frontend.dockerfile import CONTAINER_SUPPORTED_DISTROS
from bentoml.exceptions import BentoMLConfigException
from bentoml.exceptions import BentoMLException

if t.TYPE_CHECKING:
    from fs.base import FS

    from bentoml._internal.bento.build_config import BentoEnvSchema

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

logger = logging.getLogger("bentoml.build")

DEFAULT_PYTHON_VERSION = f"{sys.version_info.major}.{sys.version_info.minor}"


@attrs.define
class Image:
    """A class defining the environment requirements for bento."""

    base_image: str = ""
    distro: str = "debian"
    python_version: str = DEFAULT_PYTHON_VERSION
    commands: t.List[str] = attrs.field(factory=list)
    lock_python_packages: bool = True
    pack_git_packages: bool = True
    python_requirements: str = ""
    post_commands: t.List[str] = attrs.field(factory=list)
    scripts: t.Dict[str, str] = attrs.field(factory=dict, init=False)
    _after_pip_install: bool = attrs.field(init=False, default=False, repr=False)
    _uv_lock: bool = attrs.field(init=False, default=False, repr=False)

    def __attrs_post_init__(self) -> None:
        if not self.base_image:
            if self.distro not in CONTAINER_METADATA:
                raise BentoMLConfigException(
                    f"Unsupported distro: {self.distro}, expected one of {CONTAINER_SUPPORTED_DISTROS}"
                )
            python_version = self.python_version
            if self.distro in ("ubi8",):
                python_version = python_version.replace(".", "")
            distro_config = CONTAINER_METADATA[self.distro]
            self.base_image = distro_config["python"]["image"].format(
                spec_version=python_version
            )
        if self.distro in CONTAINER_METADATA:
            self.commands.insert(
                0, CONTAINER_METADATA[self.distro]["default_install_command"]
            )

    def system_packages(self, *packages: str) -> t.Self:
        if self.distro not in CONTAINER_METADATA:
            raise BentoMLConfigException(
                f"Unsupported distro: {self.distro}, expected one of {CONTAINER_SUPPORTED_DISTROS}"
            )
        logger.info(
            "Adding system packages using distro %s's package manager", self.distro
        )
        self.commands.append(
            CONTAINER_METADATA[self.distro]["install_command"].format(
                packages=" ".join(packages)
            )
        )
        return self

    def requirements_file(self, file_path: str) -> t.Self:
        """Add a requirements file to the image. Supports chaining call.

        Example:

        .. code-block:: python

            image = Image("debian:latest").requirements_file("requirements.txt")
        """
        if self.post_commands:
            raise BentoMLConfigException("Can't separate adding python requirements")
        self.python_requirements += Path(file_path).read_text().rstrip("\n") + "\n"
        self._after_pip_install = True
        return self

    def pyproject_toml(self, file_path: str = "pyproject.toml") -> t.Self:
        """Add a pyproject.toml to the image. Supports chaining call.

        Example:

        .. code-block:: python

            image = Image("debian:latest").pyproject_toml("pyproject.toml")
        """
        if self.post_commands:
            raise BentoMLConfigException("Can't separate adding python requirements")
        with Path(file_path).open("rb") as f:
            pyproject_toml = tomllib.load(f)
        dependencies = pyproject_toml.get("project", {}).get("dependencies", {})
        self.python_packages(*dependencies)
        return self

    def uv_lock(self) -> t.Self:
        self._uv_lock = True
        self.lock_python_packages = False  # skip locking if uv.lock is used
        return self

    def python_packages(self, *packages: str) -> t.Self:
        """Add python dependencies to the image. Supports chaining call.

        Example:

        .. code-block:: python

            image = Image("debian:latest")\
                .python_packages("numpy", "pandas")\
                .requirements_file("requirements.txt")
        """
        if self.post_commands:
            raise BentoMLConfigException("Can't separate adding python requirements")
        if not packages:
            raise BentoMLConfigException("No packages provided")
        self.python_requirements += "\n".join(packages) + "\n"
        self._after_pip_install = True
        return self

    def run(self, command: str) -> t.Self:
        """Add a command to the image. Supports chaining call.

        Example:

        .. code-block:: python

            image = Image("debian:latest").run("echo 'Hello, World!'")
        """
        commands = self.post_commands if self._after_pip_install else self.commands
        commands.append(command)
        return self

    def run_script(self, script: str) -> t.Self:
        """Run a script in the image. Supports chaining call.

        Example:

        .. code-block:: python

            image = Image("debian:latest").run_script("script.sh")
        """
        commands = self.post_commands if self._after_pip_install else self.commands
        script = Path(script).resolve().as_posix()
        # Files under /env/docker will be copied into the env image layer
        target_script = (
            f"./env/docker/script__{hashlib.md5(script.encode()).hexdigest()}"
        )
        commands.append(f"chmod +x {target_script} && {target_script}")
        self.scripts[script] = target_script
        return self

    def freeze(
        self, bento_fs: FS, envs: list[BentoEnvSchema], platform_: str | None = None
    ) -> ImageInfo:
        """Freeze the image to an ImageInfo object for build."""
        python_requirements = self._freeze_python_requirements(bento_fs, platform_)
        from importlib import resources

        from _bentoml_impl.docker import generate_dockerfile
        from bentoml._internal.utils.filesystem import copy_file_to_fs_folder

        # Prepare env/python files
        py_folder = fs.path.join("env", "python")
        bento_fs.makedirs(py_folder, recreate=True)
        reqs_txt = fs.path.join(py_folder, "requirements.txt")
        bento_fs.writetext(reqs_txt, python_requirements)
        info = ImageInfo(
            base_image=self.base_image,
            python_version=self.python_version,
            commands=self.commands,
            python_requirements=python_requirements,
            post_commands=self.post_commands,
        )
        # Prepare env/docker files
        docker_folder = fs.path.join("env", "docker")
        bento_fs.makedirs(docker_folder, recreate=True)
        dockerfile_path = fs.path.join(docker_folder, "Dockerfile")
        bento_fs.writetext(
            dockerfile_path,
            generate_dockerfile(info, bento_fs, enable_buildkit=False, envs=envs),
        )
        for script_name, target_path in self.scripts.items():
            copy_file_to_fs_folder(script_name, bento_fs, dst_filename=target_path)

        with resources.path(
            "bentoml._internal.container.frontend.dockerfile", "entrypoint.sh"
        ) as entrypoint_path:
            copy_file_to_fs_folder(str(entrypoint_path), bento_fs, docker_folder)
        return info

    def _freeze_python_requirements(
        self, bento_fs: FS, platform_: str | None = None
    ) -> str:
        from pip_requirements_parser import RequirementsFile

        from bentoml._internal.bento.bentoml_builder import build_bentoml_sdist
        from bentoml._internal.bento.build_config import PythonOptions
        from bentoml._internal.configuration import get_uv_command

        py_folder = fs.path.join("env", "python")
        bento_fs.makedirs(py_folder, recreate=True)
        requirements_in = Path(
            bento_fs.getsyspath(fs.path.join(py_folder, "requirements.in"))
        )
        uv_cmd = get_uv_command()
        with requirements_in.open("w") as f:
            if self._uv_lock:
                uv_reqs = subprocess.check_output(
                    [
                        *uv_cmd,
                        "export",
                        "--no-header",
                        "--frozen",
                        "--no-annotate",
                        "--no-hashes",
                        "--no-emit-project",
                        *(["--verbose"] if get_debug_mode() else []),
                    ],
                    text=True,
                    stderr=subprocess.DEVNULL if get_quiet_mode() else None,
                    cwd=bento_fs.getsyspath("src"),
                )
                f.write(uv_reqs.rstrip("\n") + "\n")
            f.write(self.python_requirements)
        py_req = fs.path.join("env", "python", "requirements.txt")
        requirements_out = Path(bento_fs.getsyspath(py_req))
        # XXX: RequirementsFile.from_string() does not work due to bugs
        requirements_file = RequirementsFile.from_file(str(requirements_in))
        has_bentoml_req = any(
            req.name and req.name.lower() == "bentoml" and req.link is not None
            for req in requirements_file.requirements
        )
        src_wheels = fs.path.join("src", "wheels")
        wheels_folder = fs.path.join("env", "python", "wheels")
        if bento_fs.exists(src_wheels):
            bento_fs.movedir(src_wheels, wheels_folder, create=True)
        with requirements_in.open("w") as f:
            f.write(requirements_file.dumps(preserve_one_empty_line=True))
            if not has_bentoml_req:
                sdist_name = build_bentoml_sdist(bento_fs.getsyspath(wheels_folder))
                bento_req = get_bentoml_requirement()
                if bento_req is not None:
                    logger.info(
                        "Adding BentoML requirement to the image: %s.", bento_req
                    )
                    f.write(f"{bento_req}\n")
                elif sdist_name is not None:
                    f.write(f"./wheels/{sdist_name}\n")
        if not self.lock_python_packages:
            requirements_out.parent.mkdir(parents=True, exist_ok=True)
            requirements_out.write_text(requirements_in.read_text())
            PythonOptions.fix_dep_urls(
                str(requirements_out),
                bento_fs.getsyspath(wheels_folder),
                self.pack_git_packages,
            )
            return requirements_out.read_text()
        lock_args = [
            str(requirements_in),
            "--no-header",
            f"--output-file={requirements_out}",
            "--emit-index-url",
            "--emit-find-links",
            "--no-annotate",
        ]
        if get_debug_mode():
            lock_args.append("--verbose")
        else:
            lock_args.append("--quiet")
        logger.info("Locking PyPI package versions.")
        if platform_:
            lock_args.extend(["--python-platform", platform_])
        elif platform.system() != "Linux" or platform.machine() != "x86_64":
            logger.info(
                "Locking packages for %s. Pass `--platform` option to specify the platform.",
                DEFAULT_LOCK_PLATFORM,
            )
            lock_args.extend(["--python-platform", DEFAULT_LOCK_PLATFORM])
        cmd = [*uv_cmd, "pip", "compile", *lock_args]
        try:
            subprocess.check_call(
                cmd,
                text=True,
                stderr=subprocess.DEVNULL if get_quiet_mode() else None,
                cwd=bento_fs.getsyspath(py_folder),
            )
        except subprocess.CalledProcessError as e:
            raise BentoMLException(
                "Failed to lock PyPI packages. Add `--debug` option to see more details.\n"
                "You see this error because you set `lock_packages=true` in the image config.\n"
                "Learn more at https://docs.bentoml.com/en/latest/reference/bentoml/bento-build-options.html#pypi-package-locking"
            ) from e

        PythonOptions.fix_dep_urls(
            str(requirements_out),
            bento_fs.getsyspath(wheels_folder),
            self.pack_git_packages,
        )
        return requirements_out.read_text()


def populate_image_from_build_config(
    image: Image | None, build_config: BentoBuildConfig, build_ctx: str
) -> Image | None:
    from bentoml._internal.utils.filesystem import resolve_user_filepath

    fallback_message = "fallback to bento v1" if image is None else "it will be ignored"
    if not build_config.conda.is_empty():
        logger.warning(
            "conda options are not supported by bento v2, %s", fallback_message
        )
        if image is None:
            return None
    docker_options = build_config.docker
    if docker_options.cuda_version is not None:
        logger.warning(
            "docker.cuda_version is not supported by bento v2, %s", fallback_message
        )
        if image is None:
            return None
    if docker_options.dockerfile_template is not None:
        logger.warning(
            "docker.dockerfile_template is not supported by bento v2, %s",
            fallback_message,
        )
        if image is None:
            return None

    image_params = {}
    if docker_options.base_image is not None:
        image_params["base_image"] = docker_options.base_image
    if docker_options.distro is not None:
        image_params["distro"] = docker_options.distro
    if docker_options.python_version is not None:
        image_params["python_version"] = docker_options.python_version

    python_options = build_config.python
    if python_options.wheels:
        logger.warning(
            "python.wheels is not supported by bento v2, %s\nAdd the wheels directly into your project instead.",
            fallback_message,
        )
        if image is None:
            return None
    if python_options.lock_packages is not None:
        image_params["lock_python_packages"] = python_options.lock_packages
    if python_options.pack_git_packages is not None:
        image_params["pack_git_packages"] = python_options.pack_git_packages
    image = (
        Image(**image_params) if image is None else attrs.evolve(image, **image_params)
    )
    if docker_options.system_packages:
        image.system_packages(*docker_options.system_packages)
    if python_options.index_url:
        image.python_packages(f"--index-url {python_options.index_url}")
    if python_options.no_index:
        image.python_packages("--no-index")
    if python_options.trusted_host:
        image.python_packages(
            *(f"--trusted-host {h}" for h in python_options.trusted_host)
        )
    if python_options.extra_index_url:
        image.python_packages(
            *(f"--extra-index-url {url}" for url in python_options.extra_index_url)
        )
    if python_options.find_links:
        image.python_packages(
            *(f"--find-links {link}" for link in python_options.find_links)
        )
    if python_options.requirements_txt:
        image.requirements_file(
            resolve_user_filepath(python_options.requirements_txt, build_ctx)
        )
    if python_options.packages:
        image.python_packages(*python_options.packages)
    if docker_options.setup_script:
        image.run_script(docker_options.setup_script)
    return image



from typing import TYPE_CHECKING
from typing import Any

from _bentoml_sdk.images import Image
from bentoml._internal.utils import warn_deprecated

if TYPE_CHECKING:
    PythonImage = Image


def __getattr__(name: str) -> Any:
    if name == "PythonImage":
        warn_deprecated("PythonImage is an alias of Image, use Image instead")
        return Image
    raise AttributeError(f"module {__name__} has no attribute {name}")


__all__ = ["Image"]



from __future__ import annotations

import importlib
import logging
import os
import sys
import typing as t
from typing import TYPE_CHECKING

import fs
from simple_di import Provide
from simple_di import inject

from ...exceptions import ImportServiceError
from ...exceptions import NotFound
from ..bento import Bento
from ..bento.bento import BENTO_PROJECT_DIR_NAME
from ..bento.bento import BENTO_YAML_FILENAME
from ..bento.bento import DEFAULT_BENTO_BUILD_FILES
from ..bento.build_config import BentoBuildConfig
from ..configuration import BENTOML_VERSION
from ..configuration.containers import BentoMLContainer
from ..models import ModelStore
from ..tag import Tag
from ..utils.args import set_arguments
from ..utils.uri import encode_path_for_uri

if TYPE_CHECKING:
    from _bentoml_sdk import Service as NewService

    from ..bento import BentoStore
    from .service import Service

    AnyService = t.Union[Service, NewService[t.Any]]

logger = logging.getLogger(__name__)


@inject
def import_service(
    svc_import_path: str,
    *,
    working_dir: t.Optional[str] = None,
    reload: bool = False,
    standalone_load: bool = False,
    model_store: ModelStore = Provide[BentoMLContainer.model_store],
) -> AnyService:
    """Import a Service instance from source code, by providing the svc_import_path
    which represents the module where the Service instance is created and optionally
    what attribute can be used to access this Service instance in that module

    Example usage:
        # When multiple service defined in the same module
        import_service("fraud_detector:svc_a")
        import_service("fraud_detector:svc_b")

        # Find svc by Python module name or file path
        import_service("fraud_detector:svc")
        import_service("fraud_detector.py:svc")
        import_service("foo.bar.fraud_detector:svc")
        import_service("./def/abc/fraud_detector.py:svc")

        # When there's only one Service instance in the target module, the attributes
        # part in the svc_import_path can be omitted
        import_service("fraud_detector.py")
        import_service("fraud_detector")
    """

    sys_path_modified = False
    prev_cwd = None
    global_model_store = BentoMLContainer.model_store.get()

    if working_dir is not None:
        prev_cwd = os.getcwd()
        working_dir = os.path.realpath(os.path.expanduser(working_dir))
        # Set cwd(current working directory) to the Bento's project directory,
        # which allows user code to read files using relative path
        os.chdir(working_dir)
    else:
        working_dir = os.getcwd()

    if working_dir not in sys.path:
        sys.path.insert(0, working_dir)
        sys_path_modified = True

    if model_store is not global_model_store:
        BentoMLContainer.model_store.set(model_store)

    def _restore(restore_model_store: bool = False) -> None:
        """Undo changes to sys.path, cwd and model store"""
        if restore_model_store and model_store is not global_model_store:
            BentoMLContainer.model_store.set(global_model_store)
        if sys_path_modified and working_dir:
            sys.path.remove(working_dir)
        if prev_cwd is not None:
            os.chdir(prev_cwd)

    try:
        svc = _do_import(svc_import_path, working_dir, reload)
    except ImportServiceError:
        _restore(True)
        raise
    else:
        if not standalone_load:
            _restore()
        return svc


def _do_import(
    svc_import_path: str, working_dir: str, reload: bool = False
) -> AnyService:
    from .service import Service

    try:
        from _bentoml_sdk import Service as NewService

    except ImportError:

        class NewService:
            pass

    logger.debug(
        'Importing service "%s" from working dir: "%s"', svc_import_path, working_dir
    )

    import_path, _, attrs_str = svc_import_path.partition(":")
    if not import_path:
        raise ImportServiceError(
            f'Invalid import target "{svc_import_path}", must format as '
            '"<module>:<attribute>" or "<module>'
        )

    if os.path.isfile(import_path):
        import_path = os.path.realpath(import_path)
        # Importing from a module file path:
        if not import_path.startswith(working_dir):
            raise ImportServiceError(
                f'Module "{import_path}" not found in working directory "{working_dir}"'
            )

        file_name, ext = os.path.splitext(import_path)
        if ext != ".py":
            raise ImportServiceError(
                f'Invalid module extension "{ext}" in target "{svc_import_path}",'
                ' the only extension acceptable here is ".py"'
            )

        # move up until no longer in a python package or in the working dir
        module_name_parts: t.List[str] = []
        path = file_name
        while True:
            path, name = os.path.split(path)
            module_name_parts.append(name)
            if (
                not os.path.exists(os.path.join(path, "__init__.py"))
                or path == working_dir
            ):
                break
        module_name = ".".join(module_name_parts[::-1])
    else:
        # Importing by module name:
        module_name = import_path

    # Import the service using the Bento's own model store
    needs_reload = module_name in sys.modules
    try:
        module = importlib.import_module(module_name, package=working_dir)
    except ImportError as e:
        from ..context import server_context

        message = f'Failed to import module "{module_name}": {e}'
        if server_context.worker_index is None and not (e.name or "").startswith(
            ("bentoml", "_bentoml_")
        ):
            message += "\nIf you are trying to import a runtime-only module, try wrapping it inside `with bentoml.importing():`"
        raise ImportServiceError(message)
    if reload and needs_reload:
        importlib.reload(module)

    if attrs_str:
        instance = module
        try:
            for attr_str in attrs_str.split("."):
                if isinstance(instance, NewService):
                    instance = instance.dependencies[attr_str].on
                else:
                    instance = getattr(instance, attr_str)
        except (AttributeError, KeyError):
            raise ImportServiceError(
                f'Attribute "{attrs_str}" not found in module "{module_name}".'
            )
    else:
        instances = [
            (k, v) for k, v in module.__dict__.items() if isinstance(v, NewService)
        ]
        dependents: set[NewService[t.Any]] = set()
        for _, svc in instances:
            for dep in svc.dependencies.values():
                if dep.on is not None:
                    dependents.add(dep.on)
        instances = [(k, v) for k, v in instances if v not in dependents]
        if not instances:  # Fallback to 1.1 service
            instances = [
                (k, v) for k, v in module.__dict__.items() if isinstance(v, Service)
            ]

        if not instances:
            raise ImportServiceError(f"No service found in the module '{module_name}'")
        elif len(instances) == 1:
            attrs_str, instance = instances[0]
        else:
            raise ImportServiceError(
                f'Multiple Service instances found in module "{module_name}", use '
                '"<module>:<svc_variable_name>" to specify the service instance or '
                "define only one service instance per python module/file"
            )

    # set import_str for retrieving the service import origin
    object.__setattr__(instance, "_import_str", f"{module_name}:{attrs_str}")
    return t.cast("AnyService", instance)


@inject
def load_bento(
    bento: str | Tag | Bento,
    reload: bool = False,
    standalone_build: bool = False,
    bento_store: "BentoStore" = Provide[BentoMLContainer.bento_store],
) -> AnyService:
    """Load a Service instance from a bento found in local bento store:

    Example usage:
        load_bento("FraudDetector:latest")
        load_bento("FraudDetector:20210709_DE14C9")
    """
    if isinstance(bento, (str, Tag)):
        bento = bento_store.get(bento)

    logger.debug(
        'Loading bento "%s" found in local store: %s',
        bento.tag,
        bento.path,
    )

    # not in validate as it's only really necessary when getting bentos from disk
    if bento.info.bentoml_version != BENTOML_VERSION:
        info_bentoml_version = bento.info.bentoml_version
        if tuple(info_bentoml_version.split(".")) > tuple(BENTOML_VERSION.split(".")):
            logger.warning(
                "%s was built with newer version of BentoML, which does not match with current running BentoML version %s",
                bento,
                BENTOML_VERSION,
            )
        else:
            logger.debug(
                "%s was built with BentoML version %s, which does not match the current BentoML version %s",
                bento,
                info_bentoml_version,
                BENTOML_VERSION,
            )
    return _load_bento(bento, reload=reload, standalone_build=standalone_build)


def load_bento_dir(
    path: str, reload: bool = False, standalone_build: bool = False
) -> AnyService:
    """Load a Service instance from a bento directory

    Example usage:
        load_bento_dir("~/bentoml/bentos/iris_classifier/4tht2icroji6zput3suqi5nl2")
    """
    bento_fs = fs.open_fs(encode_path_for_uri(path))
    bento = Bento.from_fs(bento_fs)
    logger.debug(
        'Loading bento "%s" from directory: %s',
        bento.tag,
        path,
    )
    return _load_bento(bento, reload=reload, standalone_build=standalone_build)


def _load_bento(
    bento: Bento, reload: bool = False, standalone_build: bool = False
) -> AnyService:
    # Use Bento's user project path as working directory when importing the service
    set_arguments(**bento.info.args)
    working_dir = bento.path_of(BENTO_PROJECT_DIR_NAME)

    model_store = BentoMLContainer.model_store.get()
    # read from bento's local model store if it exists and is not empty
    # This is the case when running in a container
    local_model_store = bento._model_store
    if local_model_store is not None and len(local_model_store.list()) > 0:
        model_store = local_model_store

    # Read the model aliases
    resolved_model_aliases = {
        m.alias: str(m.tag) for m in bento.info.all_models if m.alias
    }
    BentoMLContainer.model_aliases.set(resolved_model_aliases)

    svc = import_service(
        bento.info.service,
        working_dir=working_dir,
        reload=reload,
        standalone_load=standalone_build,
        model_store=model_store,
    )
    svc.on_load_bento(bento)
    return svc


def load(
    bento_identifier: str | Tag | Bento,
    working_dir: t.Optional[str] = None,
    reload: bool = False,
    standalone_load: bool = False,
) -> AnyService:
    """Load a Service instance by the bento_identifier

    Args:
        bento_identifier: target Service to import or Bento to load
        working_dir: when importing from service, set the working_dir
        standalone_load: treat target Service as standalone. This will change global
            current working directory and global model store.

    Returns:
        The loaded :obj:`bentoml.Service` instance.

    The argument ``bento_identifier`` can be one of the following forms:

    * Tag pointing to a Bento in local Bento store under `BENTOML_HOME/bentos`
    * File path to a Bento directory
    * "import_str" for loading a service instance from the `working_dir`

    Example load from Bento usage:

    .. code-block:: python

        # load from local bento store
        load("FraudDetector:latest")
        load("FraudDetector:4tht2icroji6zput")

        # load from bento directory
        load("~/bentoml/bentos/iris_classifier/4tht2icroji6zput")


    Example load from working directory by "import_str" usage:

    .. code-block:: python

        # When multiple service defined in the same module
        load("fraud_detector:svc_a")
        load("fraud_detector:svc_b")

        # Find svc by Python module name or file path
        load("fraud_detector:svc")
        load("fraud_detector.py:svc")
        load("foo.bar.fraud_detector:svc")
        load("./def/abc/fraud_detector.py:svc")

        # When there's only one Service instance in the target module, the attributes
        # part in the svc_import_path can be omitted
        load("fraud_detector.py")
        load("fraud_detector")
    """
    if isinstance(bento_identifier, (Bento, Tag)):
        # Load from local BentoStore
        return load_bento(
            bento_identifier, reload=reload, standalone_build=standalone_load
        )
    bento_path = os.path.expanduser(
        os.path.abspath(os.path.join(working_dir or ".", bento_identifier))
    )
    if os.path.isdir(bento_path):
        if os.path.isfile(
            os.path.expanduser(os.path.join(bento_path, BENTO_YAML_FILENAME))
        ):
            # Loading from path to a built Bento
            svc = load_bento_dir(
                bento_path, reload=reload, standalone_build=standalone_load
            )
            logger.info("Service loaded from Bento directory: %s", svc)
        else:
            for filename in DEFAULT_BENTO_BUILD_FILES:
                if os.path.isfile(
                    config_file := os.path.expanduser(
                        os.path.join(bento_path, filename)
                    )
                ):
                    build_config = BentoBuildConfig.from_file(config_file)
                    BentoMLContainer.model_aliases.set(build_config.model_aliases)
                    if build_config.service:
                        svc = import_service(
                            build_config.service,
                            working_dir=bento_path,
                            reload=reload,
                            standalone_load=standalone_load,
                        )
                        logger.debug(
                            "'%s' loaded from '%s': %s", svc.name, bento_path, svc
                        )
                        break
            else:
                if os.path.isfile(os.path.join(bento_path, "service.py")):
                    logger.info("Loading service from default location 'service.py'")
                    svc = import_service(
                        "service.py",
                        working_dir=bento_path,
                        reload=reload,
                        standalone_load=standalone_load,
                    )
                else:
                    raise ImportServiceError(
                        f"Failed to load bento or import service '{bento_identifier}'. "
                        f"The directory '{bento_path}' does not contain a valid bentofile.yaml or service.py."
                    )
    else:
        try:
            # Loading from service definition file, e.g. "my_service.py:svc"
            svc = import_service(
                bento_identifier,
                working_dir=working_dir,
                reload=reload,
                standalone_load=standalone_load,
            )
            logger.debug("'%s' imported from source: %s", svc.name, svc)
        except ImportServiceError as e1:
            try:
                # Loading from local bento store by tag, e.g. "iris_classifier:latest"
                svc = load_bento(
                    bento_identifier, reload=reload, standalone_build=standalone_load
                )
                logger.debug("'%s' loaded from Bento store: %s", svc.name, svc)
            except (NotFound, ImportServiceError) as e2:
                raise ImportServiceError(
                    f"Failed to load bento or import service '{bento_identifier}'.\n"
                    f"If you are attempting to import bento by tag: '{e2}'.\n"
                    f"If you are importing by python module path: '{e1}'."
                ) from e2
    return svc
